"""
A little built-in HTTP server for delivering some files (and possibly more).
"""

from __future__ import with_statement

import BaseHTTPServer
import socket
import threading
import urllib
from cStringIO import StringIO

import pkg_resources

from gavo import utils

BASE_TEMPLATE = """
	<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"'
		' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
<head>
<title>%%(title)s</title>
</head>
<body>
<h1>%%(title)s</h1>
%(preContent)s
%%(content)s
%(postContent)s
</body>
</html>
"""


ERROR_TEMPLATE = BASE_TEMPLATE%{
		"headMaterial": "",
		"preContent": "",
		"postContent": "",
	}


class Resource(object):
	"""Something that is delivered by the web server.

	Resources have the attributes
	* contentType
	* httpStatus,

	and a method open() -> file-like object
	"""
	contentType = "application/octet-stream"
	httpStatus = 200

	def open(self):
		raise NotImplementedError("%s resources cannot be delivered"%
			self.__class__.__name)


class TextResource(Resource):
	"""A resource containing unicode text.
	"""
	contentType = "text/html;charset=UTF-8"

	def __init__(self, content):
		if isinstance(content, unicode):
			self.content = content.encode("utf8")
		else:
			self.content = content
	
	def open(self):
		return StringIO(self.content)


class ErrorResource(TextResource):
	"""An error message, complete with an HTTP status code.
	"""
	def __init__(self, status, msg):
		self.httpStatus = status
		TextResource.__init__(self, ERROR_TEMPLATE%{
				"title": "tapsh http server error",
				"content": msg,
			})


class PackageFileResource(Resource):
	"""A file to be delivered as-is from our package resources.
	"""
	def __init__(self, pkgPath, contentType):
		self.pkgPath, self.contentType = pkgPath, contentType
	
	def open(self):
		return pkg_resources.resource_stream('tapsh', self.pkgPath)


class InternalServerHandler(BaseHTTPServer.BaseHTTPRequestHandler):
	"""The handler for the HTTP server.

	For now, we just have a mapping of "known resources" and do a 404 for
	all others.
	"""
	resources = {
		"/tapsh_icon.png":
			PackageFileResource("res/tapsh_icon.png", "image/png"),
		"/sampy_icon.png":
			PackageFileResource("res/sampy_icon.png", "image/png"),}

	def _serveResource(self, res):
		self.send_response(res.httpStatus)
		self.send_header("Content-type", res.contentType)
		self.end_headers()
		inF = res.open()
		utils.cat(inF, self.wfile)
		inF.close()
		self.wfile.close()

	def _getResource(self):
		try:
			return self.resources[self.path]
		except KeyError:
			return ErrorResource(404, "Resource not found")

	def do_GET(self):
		try:
			self._serveResource(self._getResource())
		except Exception, msg:
			self._serveResource(
				ErrorResource(500, "Internal error %s: %s"%(
					msg.__class__.__name__, unicode(msg))))

	def log_message(self, *args):
		pass


def getHostAddress():
	"""returns an IP address for the current host.

	It does this by trying to connect to the GAVO DC and then checking
	what address the connection goes out on.  Any failures lead to a
	fallback to localhost.
	"""
	try:
		sock = socket.socket()
		try:
			sock.settimeout(0.5)
			sock.connect(("dc.zah.uni-heidelberg.de", 80))
			return sock.getsockname()[0]
		finally:
			sock.close()
	except Exception, ex:
		return "127.0.0.1"


class QuietServer(BaseHTTPServer.HTTPServer):
	pass


def startServer(ctx):
	"""starts a thread running an http server on some free port.
	"""
	for portNumber in range(6700, 6750):
		try:
			address = ("", portNumber)
			server = QuietServer(address, InternalServerHandler)
		except socket.error:
			pass
		else:
			break
	else:
		# Looks like the system doesn't let us open a socket.
		# So far, no important functionality is affected, so keep quiet about it.
		return None

	httpThread = threading.Thread(target=server.serve_forever)
	httpThread.setDaemon(True)
	httpThread.start()

	return "http://%s:%s"%(getHostAddress(), portNumber)


if __name__=="__main__":
	startServer(None)
