"""
Metadata lists -- servers, jobs, miscellaneous.
"""

from __future__ import with_statement

import datetime
import re
import os
import random
import urllib
import urlparse

from gavo import utils
from gavo import votable

from tapsh import common
from tapsh import dbinter

GLOTS_SERVERS = (common.GLOTS_BASE+"/servers/form?__nevow_form__=genForm"
	"&_DBOPTIONS_LIMIT=10000&_FORMAT=VOTable")


class ServerTable(dbinter.TableDef):
	"""the table to maintain the persistent server list.

	(the context keeps an in-memory copy of this).
	"""
	c_000_ivoid = 'TEXT'
	c_010_accessURL = 'TEXT'
_servers = ServerTable("servers")


class StateTable(dbinter.TableDef):
	"""a table just keeping key-value pairs for tapsh state management.
	"""
	c_000_key = 'TEXT'
	c_010_value = 'TEXT'
_state = StateTable("state")


class JobsTable(dbinter.TableDef):
	"""the table of jobs currently "under surveillance".
	"""
	c_000_nickname = 'TEXT PRIMARY KEY' # local nickname (locally unique!)
	c_010_jobId = 'TEXT NOT NULL'       # remote job id
	c_020_endpointURL = 'TEXT NOT NULL' # services's URL
	c_030_lastphase = 'TEXT NOT NULL'   # last known phase
	c_040_lastupdate = 'TIMESTAMP NOT NULL'  # last update of the phase

	@classmethod
	def formatRec(cls, rec):
		nickname, jobId, endpointURL, lastphase, lastUpdate = rec
		parsed = urlparse.urlparse(endpointURL)
		return "%-10s %-8s %-20s %-10s %8s"%(
			nickname[:9], jobId[:7], parsed.netloc[:19],
			lastphase[:9], lastUpdate.date().isoformat())

_jobs = JobsTable("jobs")


def ensureTables():
	conn = dbinter.getConnection(common.DSN)
	dbinter.ensureTables(conn,
		utils.iterDerivedObjects(dbinter.TableDef, globals().values()))
	conn.commit()
	conn.close()


def getServerListFromNet():
	f = urllib.urlopen(GLOTS_SERVERS)
	data, metadata = votable.load(f)
	f.close()
	return [(row[0], row[1]) for row in data]


def refreshServerList(conn):
	with common.progress("Refreshing TAP server list from GAVO Data Center..."):
		serverList = getServerListFromNet()
		conn.executemany("INSERT INTO servers (ivoid, accessURL)"
			" VALUES (?,?)", serverList)


def getURLForServer(conn, ivoId):
	"""returns the accessURL of the server with ivoId (or raises a KeyError).
	"""
	res = list(conn.execute("SELECT accessURL FROM servers WHERE ivoid=?",
		(ivoId,)))
	if not res:
		raise KeyError(ivoId)
	return res[0][0]


def getServersForPrefix(conn, prefix):
	return list(conn.execute("SELECT ivoId FROM servers WHERE ivoid LIKE ?",
		(prefix+'%',)))


def getStateItem(conn, key):
	res = list(conn.execute("SELECT value FROM state WHERE key=?", (key,)))
	if res:
		return res[0][0]
	else:
		return None


def setStateItem(conn, key, value):
	cursor = conn.cursor()
	cursor.execute("DELETE FROM state WHERE key=?", (key,))
	cursor.execute("INSERT INTO state (key, value) VALUES (?,?)",
		(key, value))


_syllables = [c+v for v in 'aeiou' for c in 'bcdfglmnprstvwxz']
def makeNick():
	"""returns a random, hopefully cute-sounding nickname.

	There are 512000 nicks this function currently generates.  This may
	be a bit tight if people run many jobs and never clean up.  Hm...
	"""
	return "".join([random.choice(_syllables),
		random.choice(_syllables), random.choice(_syllables)])


def registerJob(ctx, job):
	"""registers an ADQLTAPJob in the jobs table.

	The job object receives an additional attribute, nickname.
	"""
	cursor = ctx.conn.cursor()
	while True:
		nick = makeNick()
		try:
			cursor.execute("INSERT INTO jobs (nickname, jobId, endpointURL,"
				" lastphase, lastupdate) VALUES (?, ?, ?, ?, ?)",
				(nick, job.jobId, job.endpointURL, 'PENDING', 
					datetime.datetime.utcnow()))
		except dbinter.IntegrityError:  # nick already taken, try next
			pass
		else:  # nick has been free
			break
	ctx.conn.commit()
	job.nickname = nick
	job.lastPhase = "PENDING"
	return job


def getJobByNick(ctx, nick):
	"""returns an ADQLTAPJob object for a nickname.

	The function raises a KeyError if the nickname is unknown.
	"""
	res = list(ctx.conn.execute(
		"SELECT nickname, jobid, endpointURL, lastPhase"
			" FROM jobs WHERE nickname=?", (nick,)))
	if not res:
		raise KeyError(nick)
	nickname, jobId, endpointURL, lastPhase = res[0]
	job = votable.ADQLTAPJob(endpointURL, jobId=jobId)
	job.nickname = nickname
	job.lastPhase = lastPhase
	return job


def getNicksInPhases(ctx, phaseSet):
	# I'm not bothering with a set adapter for pysqlite -- phases are
	# controlled vocabulary anyway
	nonAlphas = re.compile("[^A-Z]+")
	setLiteral = '(%s)'%(
		", ".join("'%s'"%nonAlphas.sub("", p) for p in phaseSet))
	return [row[0] for row in ctx.conn.execute(
		"SELECT nickname FROM jobs WHERE lastphase IN %s"%setLiteral)]


def updateJob(ctx, job, phase):
	"""updates job's job table entry to be phase.
	"""
	ctx.conn.execute(
		"UPDATE jobs SET lastphase=?, lastUpdate=? WHERE nickname=?",
		(phase, datetime.datetime.utcnow(), job.nickname))
	ctx.conn.commit()


def removeJob(ctx, job):
	"""removes the job from our tables (but does not otherwise touch it).
	"""
	ctx.conn.execute("DELETE FROM jobs WHERE nickname=?", (job.nickname,))
	ctx.conn.commit()


def getNicksWithPrefix(ctx, prefix):
	"""returns a list of job nicks starting with prefix.
	"""
	return list(
		row[0] for row in ctx.conn.execute(
			"SELECT nickname FROM jobs WHERE nickname LIKE ? ORDER BY nickname",
			(prefix+'%',)))
